/*
 * Enemies.cpp
 */
 
#include "enemies.h"
#include "shot.h"
#include "gameoptions.h"
#include <cstdlib>

static const Component WakeDistance = 8;
static const int MinThinkTime = 150;
static const Component SmallDistance =  0.4;
static const int AttackRecoveryTime = 100;
static const Component MinRangedAttack = 5.0;
static const Component MaxRangedAttack = 15.0;
// Distance from which to engage
static const Component MeleeDistance = 1.0;
// Spread of the melee damage (small to avoid friendlies)
static const Component MeleeRadius = 0.1;
static const Component MeleeReach = 0.5;

void Enemy::runController()
{
	if (awake) {
		// Do awake controller things
		if (attackTime >0) {
			attackTime --;
		}
		if (thinkTime > 0) {
			thinkTime --; 
		} else {
			think();
		}
		if (wallHitTime > 0) {
			wallHitTime --;
		}
		
	} else {
		// Not awake - decide whether we should wake up.
		MapBlock & block = 
			manager->map.blockAt((int) location.x, (int) location.y, layer);
		if (block.visible) {
			// Work out distance to player.
			Vect3d relPos = manager->getCentre() - location;
			Component wakeDist = WakeDistance;
			// If it's more than easy, double the wake distance.
			if (gameOptions.difficulty > 1) wakeDist *= 2;
			if (relPos.magnitudeSquared() < (WakeDistance * WakeDistance)) {
				// awaken!
				awake = true;
				initThinkTime();
				// TODO: make an appropriately horrible sound
			}
		}
	}
}

void Enemy::think()
{
	if (attackTime == 0) {
		// Determine whether to fire a ranged shot (fairies only)
		// Charge towards him
		// Or try a melee attack.
		// Charge towards player.
		targetLocation = manager->getCentre();
		Vect3d toTarget = targetLocation - location;
		Component dist = toTarget.magnitude();
		if (hasRangedAttack) {
			if (dist < MinRangedAttack) {
				fleeTarget();
			} else {
				if (dist > MaxRangedAttack) {
					seekTarget();
				} else {
					attackTime = MinThinkTime;
					if (gameOptions.difficulty > 2) 
						attackTime = attackTime / 2;
					doAttack();
				}
			}
		} else {
			if (dist > MeleeDistance) {
				seekTarget();
			} else {
				attackTime = MinThinkTime;
				if (gameOptions.difficulty > 2) 
					attackTime = attackTime / 2;
				doAttack();
			}
		}
		initThinkTime();
	}
}

void Enemy::initThinkTime()
{
	thinkTime = MinThinkTime;
	thinkTime += (std::rand() % 50);
	// Brown trousers mode - more thinking.
	if (gameOptions.difficulty > 2) 
		thinkTime = thinkTime / 2;
	
}

void Enemy::seekTarget()
{
	Vect3d toTarget = targetLocation - location;
	// Set desiredSpeed.
	desiredSpeed = maxSpeed;
	Component targetDistance = toTarget.magnitude();
	// Set desired Azimuth.
	if (targetDistance > SmallDistance) {
		// Target is significantly far away.
		desiredAzimuth = toTarget.angle();
	} else {
		// Arrived at target.
		thinkTime = 0;
	}
}

void Enemy::fleeTarget()
{
	Vect3d toTarget = targetLocation - location;
	// Set desiredSpeed.
	desiredSpeed = maxSpeed;
	// Set desired Azimuth.
	Vect3d fromTarget = - (toTarget);
	desiredAzimuth = fromTarget.angle();
}

void Enemy::hitWall(int blockx, int blocky)
{
	// Charge off in a random direction instead.
	// Because we might hit another wall very soon, wait a while
	if (wallHitTime == 0) {
		desiredAzimuth = (std::rand() % 360) - 180;
		initThinkTime();
		wallHitTime = 100;
	}
}

Enemy::Enemy(Manager *m) : GameCharacter(m) {
	awake = false;
	wallHitTime = 0;
	meleeDamage = 5;
	attackTime = 0;
}

Enemy::~Enemy() {
}

void Enemy::handleCollision(const Vect3d & overlap, GameObject *what)
{
	GameCharacter::handleCollision(overlap, what);
	think();
}


void EnemyKnight::doAttack()
{
	desiredSpeed = 0;
	// Set the target azimuth.
	Vect3d toTarget = targetLocation - location;
	desiredAzimuth = toTarget.angle();
	// Work out whether we're facing the right direction...
	Component adiff = desiredAzimuth - azimuth;
	normaliseAngle(adiff);
	if (fabs(adiff) < 30) {
		// If so, hit.
		// Otherwise, wait until we are.
		Component damage = meleeDamage + (std::rand() % 6);
		Vect3d meleeTarget(location + Vect3d::fromPolar(azimuth, MeleeReach));
		manager->areaDamage(meleeTarget, damage, MeleeRadius, this);
	}
}

void EnemyFairy::doAttack()
{
	desiredSpeed = 0;
	// Set the target azimuth.
	Vect3d toTarget = targetLocation - location;
	desiredAzimuth = toTarget.angle();
	// Work out whether we're facing the right direction...
	Component adiff = desiredAzimuth - azimuth;
	normaliseAngle(adiff);
	if (fabs(adiff) < 30) {
		// Fire off a round.
		float shotSpeed = 4.0;
		// Brown trousers mode? shots double speed.
		if (gameOptions.difficulty > 2) 
			shotSpeed *= 2;
		Shot *shot = new Shot(manager , location, Vect3d::fromPolar(azimuth, shotSpeed));
		shot->setOwner(this);
		manager->addObject(shot);
	}
}
